/*
    作者：裴承轩
    原理：利用并查集并路径压缩，保证建树的深度就只有2，并添加一个头节点，一个尾节点，所有第一排打开的节点的父节点都是这个头节点或者这个头节点的根节点，所有最后一排的节点的
    父节点都是那个尾节点或者尾节点的根节点，其他节点判断周围的点是否打开，打开的话就将其加入这个点的并查集，高度压缩在搜寻父节点的时候完成。father数组用来存储父节点的下表
    

*/
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include <cstring>

int find_father(int index, int *father)
{
    if (*(father + index) != index)
        *(father + index) = find_father(*(father + index), father);
    return *(father + index);
}

int get_index(int i, int j, int N)
{
    return i * N + j;
}

int main(int argc, char **args)
{
    clock_t start = clock();

    srand(time(0));

    int N = atoi(args[1]);
    int T = atoi(args[2]);

    int size = N * N;
    int times = T;

    double percentage = 1.0 / size;

    double mean = 0;
    double stddev = 0;
    double confidenceLo = 0;
    double confidenceHi = 0;

    double *answer = (double *)malloc(sizeof(double) * T);

    for (int i = 0; i < T; ++i)
        *(answer + i) = 0;

    int *father = (int *)malloc(sizeof(int) * size + 2); //one for head, one for tail
    int *status = (int *)malloc(sizeof(int) * size + 2); //0 for blocked, 1 for open.

    while (T--)
    {
        bool percolates = false;

        for (int i = 0; i < size + 2; ++i)
        {
            *(father + i) = i;
        }

        for (int i = 0; i < size + 2; ++i)
        {
            *(status + i) = 0;
        }

        while (!percolates)
        {
            int i = rand() % N;
            int j = rand() % N;

            int index = get_index(i, j, N);

            if (*(status + index) == 0)
            {
                *(status + index) = 1;

                if (N != 2)
                {
                    if (i == 0) //如果size为2*2的话，就不能使用一头一尾加一个节点的判断方法，会死循环
                    {
                        *(father + index) = find_father(size, father);
                        if(*(status + index + N))
                        {
                            *(father + find_father(index + N, father)) = find_father(index, father);
                        }
                    }
                    else if (i == N - 1)
                    {
                        *(father + index) = find_father(size + 1, father);
                        if(*(status + index - N))
                        {
                            *(father + find_father(index - N, father)) = find_father(index, father);
                        }
                    }
                    else
                    {
                        if (j != 0 && *(status + index - 1)) //left
                        {
                            *(father + find_father(index - 1, father)) = *(father + find_father(index, father));
                        }

                        if (j != N - 1 && *(status + index + 1)) //right
                        {
                            *(father + find_father(index + 1, father)) = *(father + find_father(index, father));
                        }

                        if (i != N - 1 && *(status + index + N)) //down
                        {
                            *(father + find_father(index + N, father)) = *(father + find_father(index, father));
                        }

                        if (i != 0 && *(status + index - N)) //up
                        {
                            *(father + find_father(index - N, father)) = *(father + find_father(index, father));
                        }
                    }
                }
                *(answer + T) += percentage;
            }

            if (find_father(size, father) == find_father(size + 1, father))
            {
                percolates = true;
            }

            if (N == 2)
            {
                if ((*(status) && *(status + 2)) || (*(status + 1) && *(status + 3)))
                {
                    percolates = true;
                }
            }
        }

        percolates = false;
    }

    for (int i = 0; i < times; ++i)
    {
        mean += (*(answer + i));
    }

    mean /= times;

    for (int i = 0; i < times; ++i)
    {
        stddev += pow((*(answer + i) - mean), 2);
    }

    stddev /= (times - 1);

    confidenceLo = mean - 1.96 * sqrt(stddev) / sqrt(times);
    confidenceHi = mean + 1.96 * sqrt(stddev) / sqrt(times);

    clock_t finish = clock();

    printf("mean         = %f\n", mean);
    printf("stddev       = %f\n", sqrt(stddev));
    printf("confidenceLo = %f\n", confidenceLo);
    printf("confidenceHi = %f\n", confidenceHi);
    printf("time         = %fs\n", (double)(finish - start) / CLOCKS_PER_SEC);

    return 0;
}
