#include <iostream>
#include <cstdlib>
#include <ctime>

using namespace std;

void merge(int *aux, int l, int m, int r)
{
    int size_1 = m - l + 1;
    int size_2 = r - m;
    int a[size_1];
    int b[size_2];
    int x = 0;
    int y = 0;

    for (int i = l; i <= m; ++i)
    {
        a[x++] = aux[i];
    }

    for (int j = m + 1; j <= r; ++j)
    {
        b[y++] = aux[j];
    }

    x = 0;
    y = 0;

    while (x < size_1 && y < size_2)
    {
        if (a[x] > b[y])
            aux[l++] = a[x++];
        else
            aux[l++] = b[y++];
    }

    if (x == size_1)
        while (l <= r)
            aux[l++] = b[y++];
    else if (y == size_2)
        while (l <= r)
            aux[l++] = a[x++];
}

void tdm_mergesort(int *aux, int l, int r)
{
    int m = (l + r) / 2;
    if (l >= r)
        return;

    tdm_mergesort(aux, l, m);
    tdm_mergesort(aux, m + 1, r);
    merge(aux, l, m, r);
}

void bum_mergesort(int *aux, int l, int r)
{
    auto less = [=](int a, int b) {if (a > b) return b; else return a; };
    for (int sz = 1; sz <= r; sz *= 2)
        for (int i = 0; i < r - sz + 1; i += 2 * sz)
            merge(aux, i, i + sz - 1, less(r, i + 2 * sz - 1));
}

void exchange(int &a, int &b)
{
    int tmp = a;
    a = b;
    b = tmp;
}

void quick_sort(int *aux, int l, int r)
{
    if (r <= l)
        return;

    int tmp = aux[l];
    int i = l;
    int j = r + 1;
    while (true)
    {
        while (tmp < aux[--j] && j != l)
            ;
        while (tmp > aux[++i] && i != r)
            ;
        if (j <= i)
            break;
        exchange(aux[i], aux[j]);
    }
    exchange(aux[l], aux[j]);

    quick_sort(aux, l, j - 1);
    quick_sort(aux, j + 1, r);
}

void insertion_sort(int *num, int length)
{
    for (int i = 1; i < length; ++i)
        for (int j = i - 1; j >= 0 && num[j + 1] > num[j]; --j)
            exchange(num[j + 1], num[j]);
}

void three_way_quick_sort(int *aux, int lo, int hi)
{
    if (hi <= lo)
        return;

    int lt = lo;
    int i = lo + 1;
    int gt = hi;
    int tmp = aux[lo];
    while (i <= gt)
    {
        if (aux[i] < tmp)
        {
            exchange(aux[lt++], aux[i++]);
        }
        else if (aux[i] > tmp)
            exchange(aux[i], aux[gt--]);
        else
            i++;
    }

    three_way_quick_sort(aux, lo, lt - 1);
    three_way_quick_sort(aux, gt + 1, hi);
}

int main()
{
    int size;
    int range;

    srand(time(0));

    cout << "Please input size: ";
    cin >> size;
    cout << "Please input range: ";
    cin >> range;

    int *aux = (int *)malloc(sizeof(int) * size);
    int *aux_1 = (int *)malloc(sizeof(int) * size);
    int *aux_2 = (int *)malloc(sizeof(int) * size);
    int *aux_3 = (int *)malloc(sizeof(int) * size);
    int *aux_4 = (int *)malloc(sizeof(int) * size);
    int *aux_5 = (int *)malloc(sizeof(int) * size);

    for (int i = 0; i < size; ++i)
    {
        aux[i] = rand() % range;
        aux_1[i] = aux[i];
        aux_2[i] = aux[i];
        aux_3[i] = aux[i];
        aux_4[i] = aux[i];
        aux_5[i] = aux[i];
    }

    clock_t zero = clock();
    insertion_sort(aux_1, size);
    clock_t one = clock();
    cout << "insertion_sort time:       " << (double)(one - zero) / CLOCKS_PER_SEC << endl;

    clock_t two = clock();
    tdm_mergesort(aux_2, 0, size - 1);
    clock_t three = clock();
    cout << "tdm_mergesort time:        " << (double)(three - two) / CLOCKS_PER_SEC << endl;

    clock_t four = clock();
    bum_mergesort(aux_3, 0, size - 1);
    clock_t five = clock();
    cout << "bum_mergesort time:        " << (double)(five - four) / CLOCKS_PER_SEC << endl;

    clock_t six = clock();
    quick_sort(aux_4, 0, size - 1);
    clock_t seven = clock();
    cout << "quick_sort time:           " << (double)(seven - six) / CLOCKS_PER_SEC << endl;

    clock_t eight = clock();
    three_way_quick_sort(aux_5, 0, size - 1);
    clock_t nine = clock();
    cout << "three_way_quick_sort time: " << (double)(nine - eight) / CLOCKS_PER_SEC << endl;

    return 0;
}
